import logging

from .SentioRegisterMap import *
from .ModbusWrapper import ModbusWrapper
from .SentioTypes import *
from .Defaults import Defaults

class SentioItcHandler:
    def __init__(self, api:ModbusWrapper):
        self._detectedItcs = []
        self._api = api
        pass

    def initialize(self):
        self._detectedItcs=[]
        for itcCtr in range(Defaults.MaxNumberOfItcs):
            try:
                value = self._api.readRegister(SentioRegisterMap.Itc_Name, _subIndex = itcCtr)
                if value != None:
                    self._detectedItcs.append(SentioItcCircuit(itcCtr, value, self._api))
                #else:
                    #logging.error("ITC Circuit {0} not found".format(itcCtr))
            except Exception as e:
                logging.exception("Exception reading ITC Circuit {0}".format(e))
                return -1

        #for circuit in self._detectedItcs:
        #    logging.debug("ITC Circuit ID={0} | Name={1}".format(circuit.index, circuit.name))
        return 0
    
    def updateData(self):
        for itc in self._detectedItcs:
            itc.updateData()

class SentioItcCircuit:
    def __init__(self, index, name, api:ModbusWrapper):
        self.index = index;
        self.name = name;
        self._state         = None
        self._pumpState     = None
        self._inletMeasured = None
        self._inletDesired  = None
        self._returnTemp    = None
        self._supplierTemp  = None
        self._api = api

    def GetItcIndex(self):
        return self.index
    
    @property
    def getPumpState(self) -> ITC_PumpState:
        return self._pumpState
    
    @property
    def getInletMeasured(self) -> float:
        return self._inletMeasured
    
    @property
    def getInletMeasured(self) -> float:
        return self._inletMeasured
        
    @property
    def getInletDesired(self) -> float:
        return self._inletDesired
    @property
    def getSupplierTemp(self) -> float:
        return self._supplierTemp
    
    @property
    def getReturnTemp(self) -> float:
        return self._returnTemp
    
    def updateData(self):
        state = self._api.readRegister(SentioRegisterMap.Itc_State, _subIndex = self.index)
        if state:
            self._state          = SentioHeatingStates(state)
        pump = self._api.readRegister(SentioRegisterMap.Itc_PumpState, _subIndex = self.index)
        if pump:
            self._pumpState = ITC_PumpState(pump)
        self._inletMeasured  = self._api.readRegister(SentioRegisterMap.Itc_InletTempMeasured, _subIndex = self.index)
        if(self._inletMeasured == Defaults.InvalidFP2_100):
            self._inletMeasured = None
        self._inletDesired   = (self._api.readRegister(SentioRegisterMap.Itc_InletTempDesired, _subIndex = self.index))
        if(self._inletDesired == Defaults.InvalidFP2_100):
            self._inletDesired = None
        self._returnTemp     = (self._api.readRegister(SentioRegisterMap.Itc_ReturnTemperature, _subIndex = self.index))
        if(self._returnTemp == Defaults.InvalidFP2_100):
            self._returnTemp = None
        self._supplierTemp   = (self._api.readRegister(SentioRegisterMap.Itc_SupplierTemperature, _subIndex = self.index))
        if(self._supplierTemp == Defaults.InvalidFP2_100):
            self._supplierTemp = None
        pass

    def printCircuit(self):
        logging.debug("ITC Circuit ID={0} | Name={1} || state {2}, Pump {3}, Inlet = {4} desired={5}, Return ={6}, Supplier ={7}".format(self.index, self.name, self._state, self._pumpState, self._inletMeasured, self._inletDesired, self._returnTemp, self._supplierTemp))
        print("ITC Circuit ID={0} | Name={1} || state {2}, Pump {3}, Inlet = {4} desired={5}, Return ={6}, Supplier ={7}".format(self.index, self.name, self._state, self._pumpState, self._inletMeasured, self._inletDesired, self._returnTemp, self._supplierTemp))


import logging

from pymodbus.client import ModbusTcpClient
from pymodbus.client import ModbusSerialClient
from pymodbus.payload import BinaryPayloadDecoder
from pymodbus.constants import Endian

#Api regisers and defaults
#import Defaults
from .Defaults import Defaults
from .SentioRegisterMap import *
from .SentioTypes import *
from .ModbusWrapper import ModbusWrapper
from .SentioRoom import SentioRoom, SentioRoomHandler
from .SentioITC import SentioItcCircuit, SentioItcHandler
from .SentioSensors import SentioSensors

class ModbusType(Enum):
    MODBUS_TCPIP = 0
    MODBUS_RTU = 1

class SentioModbus:
    def __init__(self, 
        _ModbusType:ModbusType = ModbusType.MODBUS_TCPIP,
        _host: str = "",
        _baudrate: int = Defaults.BaudRate,
        _slaveId: int = Defaults.SlaveId,
        _port: int = Defaults.TcpPort,
        _loglevel = logging.ERROR,
        _parity = 'E',
        _stopbits = 1
    ):
        self.host = _host
        self.slaveId = _slaveId
        self.port = _port
        self.modbusType = _ModbusType
        self.baudrate = _baudrate
        self.parity = _parity
        self.stopbits = _stopbits

        logging.basicConfig(format='%(levelname)s %(asctime)s | %(message)s', level=_loglevel, datefmt='%m/%d/%Y %H:%M:%S')
        self.detectedRooms=[]
        pass

    def connectModbusRTU(self):
        try:
            self.client = ModbusSerialClient(method='rtu', port=self.host, baudrate=self.baudrate, parity=self.parity, stopbits=self.stopbits) #fixed parity & stopbits?
            result = self.client.connect()
            if result == True:
                self.modbusWrapper = ModbusWrapper(self.client, self.slaveId)
                logging.debug("Connected {0}".format(result))
                #super().connect()
                return 0
            else:
                logging.error("Failed to connect")
                return -1
        except Exception as error:
            logging.exception("Failed to connect: {0} ".format(error))
            return -1

    def connectModbusTcpIp(self):
        try:
            self.client = ModbusTcpClient(self.host)
            logging.debug("Connecting")
            result = self.client.connect()
            if result == True:
                self.modbusWrapper = ModbusWrapper(self.client, self.slaveId)
                logging.debug("Connected {0}".format(result))
                #super().connect()
                return 0
            else:
                logging.error("Failed to connect")
                return -1
        except:
            logging.exception("Failed to connect")
            raise NoConnectionPossible("Failed to connect")

    def connect(self):
        if self.modbusType == ModbusType.MODBUS_TCPIP:
           return self.connectModbusTcpIp()
        elif self.modbusType == ModbusType.MODBUS_RTU:
            return self.connectModbusRTU()
        elif self.modbusType == 0:
            return self.connectModbusTcpIp() #ugly fallback
        else: 
            logging.exception("Wrong modbus type detected {0}".format(self.modbusType))
            raise AttributeError("Wrong modbus type detected {0}".format(self.modbusType))

    def disconnect(self):
        return self.client.close()

    def initialize(self):
        self._sentioData = SentioSensors(self.modbusWrapper)
        self._roomData = SentioRoomHandler(self.modbusWrapper)
        self._itcData = SentioItcHandler(self.modbusWrapper)

        if self._sentioData.initialize() == 0:
            if self._roomData.initialize() == 0:
                return self._itcData.initialize()
            else:
                logging.error("Failed to initialize room data")
        else:
            logging.error("Failed to initialize Sentio data")
        return -1

    def updateData(self):
        self._sentioData.updateData()
        self._roomData.updateData()
        self._itcData.updateData()
        pass

    def detectDHW(self):
        output = self.modbusWrapper.readRegister(SentioRegisterMap.DHWDeviceName)
        return output != None #if None is returned, it is not valid

    def detectCMV(self):
        output = self.modbusWrapper.readRegister(SentioRegisterMap.CMVDeviceName)
        return output != None

    def detectDehumidifiers(self):
        returnValue=[]
        for i in range(4):
            returnValue.append((self.modbusWrapper.readRegister(SentioRegisterMap.DehumDeviceName, _subIndex=i) != None))
        return returnValue

    @property
    def sentioData(self):
        return self._sentioData
    
    @property
    def availableRooms(self):
        return self._roomData._detectedRooms
    
    @property
    def availableItcs(self):
        return self._itcData._detectedItcs
    
class NoConnectionPossible(Exception):
    pass


import logging

from .SentioRegisterMap import *
from .ModbusWrapper import ModbusWrapper
from .SentioTypes import *
from .Defaults import Defaults

class SentioSensors:
    def __init__(self, api:ModbusWrapper):
        self._outdoorTemp = None
        self._hcsourceState = None
        
        self._deviceType = None
        self._serialNumber = None
        self._firmwareVersionMajor = None
        self._firmwareVersionMinor = None
        
        self._api = api
    
     ### TODO remove?
    #def readDeviceData(self):
    #    devType = self.modbusWrapper.readRegister(SentioRegisterMap.DeviceType)
    #    logging.info("DeviceType           = {0}".format(devType))
    #    logging.info("SerialNumberPrefix   = {0}".format(self.modbusWrapper.readRegister(SentioRegisterMap.DeviceSerialNrPrefix)))
    #    logging.info("SerialNumber         = {0}".format(self.modbusWrapper.readRegister(SentioRegisterMap.DeviceSerialNumber)))
    #    logging.info("FirmwareVersion         = {0}".format(self.modbusWrapper.readRegister(SentioRegisterMap.DeviceSerialNumber)))
    #    logging.info("ModbusMode           = {0}".format(self.modbusWrapper.readRegister(SentioRegisterMap.ModbusMode)))
        
    def initialize(self):
        returnValue = 0
        try:
            deviceTypeInt = self._api.readRegister(SentioRegisterMap.DeviceType)
            if deviceTypeInt:
                self._deviceType = SentioDeviceType(deviceTypeInt)
            else:
                self._deviceType = None

            serialNumberPrefix =  self._api.readRegister(SentioRegisterMap.DeviceSerialNrPrefix)
            serialNumberNumber  = self._api.readRegister(SentioRegisterMap.DeviceSerialNumber)
            serialNumberNumber = str(serialNumberNumber).zfill(Defaults.SerialNumberLength - Defaults.SerialNumberPrefixLen)
            self._serialNumber  = "{0}-{1}-{2}-{3}".format(serialNumberPrefix, serialNumberNumber[:2], serialNumberNumber[2:6], serialNumberNumber[6:10])

            self._firmwareVersionMajor = self._api.readRegister(SentioRegisterMap.DeviceSwVersion)
            self._firmwareVersionMinor = self._api.readRegister(SentioRegisterMap.DeviceSwVersionMinor)  

        except Exception as e:
            logging.error("Exception occured ==> {0}".format(e))
            returnValue = -1
        return returnValue

    def updateData(self):      
        #repetative values:
        self._outdoorTemp = self._api.readRegister(SentioRegisterMap.OutdoorTemperature)
        if self._outdoorTemp and self._outdoorTemp == Defaults.InvalidFP2_100:
            self._outdoorTemp = None
        hcstate = self._api.readRegister(SentioRegisterMap.HCSourceState)
        if hcstate:
            self._hcsourceState = SentioHeatingStates(hcstate)


    @property
    def outdoor_temperature(self) -> float:
        """Return outdoor temperature."""
        return self._outdoorTemp
    
    @property
    def hc_source_state(self) -> float:
        """Return H/C source state."""
        return self._hcsourceState

    @property
    def device_type(self) -> SentioDeviceType:
        return self._deviceType

    @property
    def serial_number(self):
        return self._serialNumber
    
    @property
    def firmware_version_major(self) -> int:
        return self._firmwareVersionMajor

    @property
    def firmware_version_minor(self) -> int:
        return self._firmwareVersionMinor

    @property
    def firmware_version_major(self) -> int:
        return self._firmwareVersionMajor


import logging

from enum import Enum

class SentioDeviceType(Enum):
    CCU208 = 1
    DHW201 = 2

class SentioCMVVentilationState(Enum):
    STOPPED = 0
    UNOCCUPIED = 1
    ECONOMY = 2
    COMFORT = 3
    BOOST = 4
    BLOCKED_STOPPED = 5
    BLOCKED_UNOCCUPIED = 6
    BLOCKED_ECONOMY = 7
    BLOCKED_COMFORT = 8
    BLOCKED_BOOST = 9
    FAILURE = 10
    MAINTENANCE = 11

class SentioHeatingStates(Enum):
    IDLE = 1
    HEATING = 2 
    COOLING = 3
    BLOCKED_HEATING = 4 
    BLOCKED_COOLING = 5

class SentioRoomMode(Enum):
    SCHEDULE = 0
    MANUAL = 1
    
class SentioRoomModeSetting(Enum):
    RM_NONE = 0
    RM_TEMPORARY = 1
    RM_VACATION_AWAY = 2
    RM_ADJUST = 3

class SentioRoomPreset(Enum):
    RP_INVALID = -1
    RP_ECO= 0
    RP_COMFORT = 1
    RP_EXTRA_COMFORT = 2


class ITC_PumpState(Enum):
    PUMP_IDLE = 1
    PUMP_ON = 2

class SentioRoomBlockingReasons(Enum):
    NONE = 0
    UNKNOWN = 1
    CONTACT = 2
    FLOOR_TEMP = 3
    LOW_ENERGY = 4
    AIR_TEMP = 5
    DEW_POINT = 6
    OUTDOOR_TEMP = 7
    FAULT = 8# (general fault, e.g. missing sensors)
    FAULT_HTCO = 9
    PERIODIC_ACTIVATION = 10
    BMS = 11
    DEADBAND = 12
    DRYING = 13
    HEATING_COOLING_MODE = 14
    INSUFFICIENT_DEMAND = 15
    COOLDOWN_PERIOD=16
    HCW_SOURCE_NOT_RELEASED = 17
    ROOM_MODE = 18
    SYSTEM_IS_INITIALIZING = 19
    SYSTEM_IS_SHUTTING_DOWN = 20
    NO_OUTPUT = 21
    FIRST_OPEN_ACTIVATION = 22
    ROOM_WITH_NO_TEMPERATURE = 23



import logging

from .SentioRegisterMap import *
from .ModbusWrapper import ModbusWrapper
from .SentioTypes import *
from .Defaults import Defaults

class SentioRoomHandler:
    def __init__(self, api:ModbusWrapper):
        self._detectedRooms = []
        self._api = api
        pass

    def initialize(self):
        roomCtr = 0
        self._detectedRooms = []
        for roomCtr in range(Defaults.MaxNumberOfRooms):
            try:
                value = self._api.readRegister(SentioRegisterMap.RoomName, _subIndex = roomCtr)
                if value != None:
                    self._detectedRooms.append(SentioRoom(roomCtr, value, self._api))
                #else:
                    #logging.error("Room {0} not found".format(roomCtr))
            except Exception as e:
                logging.exception("Exception reading room {0}".format(e))
                return -1

        for room in self._detectedRooms:
            logging.debug("Room ID={0} | Name={1}".format(room.index, room.name))
        return 0
    
    def updateData(self):
        for room in self._detectedRooms:
            room.updateData()

class SentioRoom:
    def __init__(self, index, name, api:ModbusWrapper):
        self.index = index;
        self.name = name;
        self.roomMode = SentioRoomMode.MANUAL
        self.heatingState = SentioHeatingStates.IDLE
        self.airTemp = 0.0
        self.temperatureSetPoint = 0.0
        self.humidity = 0.0
        self.floorTemp = 0.0
        self.roomBlockingMode = None
        self._api = api

    def GetRoomIndex(self):
        return self.index
    
    def GetAirTemp(self):
        return self.CurrentTemp

    def GetFloorTemp(self):
        return self.floorTemp

    def SetAirTemp(self, temperature):
        self.CurrentTemp = temperature
        
    def SetFloorTemp(self, temperature):
        self.floorTemp = temperature

    def GetFloorTemp(self):
        return self.floorTemp

    def SetTemperatureSetpoint(self, setpoint):
        self.SetPoint = setpoint

    def __str__(self):
        return "{0} - {1}".format(self.index, self.name)

    def updateData(self):
        #print("Updating room {0} data now".format(self.index))
        tempSetpoint = self._api.readRegister(SentioRegisterMap.RoomTempSetpoint, _subIndex = self.index)
        if tempSetpoint:
            self.temperatureSetPoint = tempSetpoint / 100
        else:
            self.temperatureSetPoint = None
        self.calculatedDewPoint = self._api.readRegister(SentioRegisterMap.RoomCalcDewPoint, _subIndex = self.index)
        if self.calculatedDewPoint: 
            self.calculatedDewPoint = self.calculatedDewPoint / 100
        if(self.calculatedDewPoint == Defaults.InvalidFP2_100):
            self.calculatedDewPoint = None
        self.floorTemp  = self._api.readRegister(SentioRegisterMap.RoomFloorTempActual, _subIndex = self.index)
        if self.floorTemp: 
            self.floorTemp = self.floorTemp / 100
        if(self.floorTemp == Defaults.InvalidFP2_100):
            self.floorTemp = None
        self.airTemp   = self._api.readRegister(SentioRegisterMap.RoomAirTempActual, _subIndex = self.index)
        if self.airTemp: 
            self.airTemp = self.airTemp / 100
        self.humidity  = self._api.readRegister(SentioRegisterMap.RoomRelHumidity, _subIndex = self.index)
        if self.humidity: 
            self.humidity = self.humidity / 100
        if(self.humidity == Defaults.InvalidFP2_100):
           self.humidity = None

        value = self._api.readRegister(SentioRegisterMap.RoomHeatingState, _subIndex = self.index)
        if value:
            self.heatingState = SentioHeatingStates(value)
        value = self._api.readRegister(SentioRegisterMap.RoomMode, _subIndex = self.index)
        if value:
            self.roomMode = SentioRoomMode(value)
        value = self._api.readRegister(SentioRegisterMap.RoomPresetMode, _subIndex = self.index)
        if value:
            self.roomPreset = SentioRoomPreset(value)
        if self.heatingState == SentioHeatingStates.BLOCKED_COOLING or self.heatingState == SentioHeatingStates.BLOCKED_HEATING:
            value = self._api.readRegister(SentioRegisterMap.RoomBlockingReason, _subIndex = self.index)
            if value:
                self.roomBlockingMode = SentioRoomBlockingReasons(value)
        else:
            self.roomBlockingMode = None
        pass

    def setRoomSetpoint(self, setpointdouble):
        mode = self._api.readRegister(SentioRegisterMap.RoomModeOverride, _subIndex = self.index)
        if mode and mode != SentioRoomModeSetting.RM_NONE:
            self._api.writeRegister(SentioRegisterMap.RoomModeOverride, SentioRoomModeSetting.RM_NONE.value, _subIndex = self.index)

        setpoint_fp100 = int(setpointdouble * 100)
        logging.debug(" ---------- Set room {0} temp setpoint to {1}".format(self.index, setpoint_fp100))
        value = self._api.writeRegister(SentioRegisterMap.RoomTempSetpChange, setpoint_fp100, _subIndex = self.index)
        self.temperatureSetPoint = setpointdouble
        #print("Setting room temperature setpoint to " + setpointdouble + " result " + value)
        return value
    
    def getRoomSetpoint(self):
        return self.temperatureSetPoint
    
    def getRoomActualTemperature(self):
        return self.airTemp
        
    def getRoomRelativeHumidity(self):
        return self.humidity

    def getRoomFloorTemperature(self):
        return self.floorTemp

    def getRoomCalculatedDewPoint(self):
        return self.calculatedDewPoint

    def getRoomHeatingState(self):
        return self.heatingState

    def getRoomMode(self):
        return self.roomMode

    def setRoomMode(self, roomMode:SentioRoomMode):
        value = self._api.writeRegister(SentioRegisterMap.RoomMode, roomMode.value, _subIndex = self.index)
        print("Setting Room {0} to set room {1} {2} == Result:{3}".format(self.index, roomMode, roomMode.value, value))
        self.roomMode = roomMode

    def setRoomPreset(self, roomPreset:SentioRoomPreset):
        val = self._api.writeRegister(SentioRegisterMap.RoomPresetMode, roomPreset.value, _subIndex = self.index)
        print("Setting Room Preset {0} {1}".format(roomPreset.value, val))
        self.roomPreset = roomPreset

    def setRoomHeatingState(self, roomHvacMode:SentioRoomModeSetting):
        #todo
        
        pass




import logging 
#Api regisers and defaults
from pymodbus.client import ModbusTcpClient
from pymodbus.client import ModbusSerialClient
from pymodbus.payload import BinaryPayloadDecoder
from pymodbus.constants import Endian

#import Defaults
from .Defaults import Defaults
from .SentioRegisterMap import *
from .SentioTypes import *

class ModbusWrapper:

    def __init__(self, _client, slaveId):
        self.client = _client
        self.slaveId = slaveId


    def writeRegister(self, registerMapObject, value, _subIndex = 0):
        returnValue = 0
        
        try:
            address = registerMapObject.address
            if(registerMapObject.objectType == RegisterObjectType.ROOM):
                logging.debug("Detected Room Object {0} - {1}".format(address, _subIndex))
                address = (_subIndex * 100) + address
            if(registerMapObject.regType == RegisterType.INPUT_REGISTER):
                logging.error("Not possible to write input reg")
            elif(registerMapObject.regType == RegisterType.DISCRETE_INPUT):
                logging.error("Not possible to write discrete input reg")
            elif(registerMapObject.regType == RegisterType.HOLDING_REGISTER):
                returnValue = self.client.write_register(address, value, slave=self.slaveId)
        except Exception as e:
            logging.exception("error occured ==>  {0}".format(e))
            returnValue = -1
        return returnValue


    def readRegister(self, registerMapObject, _subIndex = 0):
        returnValue = 0
        logging.debug("Reading {0}".format(registerMapObject))
        try:
            address = registerMapObject.address
            if(registerMapObject.objectType == RegisterObjectType.ROOM):
                logging.debug("Detected Room Object {0} - {1}".format(address, _subIndex))
                address = (_subIndex * 100) + address
            if(registerMapObject.objectType == RegisterObjectType.DEHUMIDIFIERS):
                logging.debug("Detected Dehumidifier Object {0} - {1}".format(address, _subIndex))
                address = (_subIndex * 100) + address
            if(registerMapObject.objectType == RegisterObjectType.ITC_CIRCUIT):
                logging.debug("Detected ITC Circuit {0} - {1}".format(address, _subIndex))
                address = (_subIndex * 100) + address


            if(registerMapObject.regType == RegisterType.INPUT_REGISTER):
                response = self.client.read_input_registers(address, registerMapObject.count, slave=self.slaveId)
                if not response.isError(): 
                    if(registerMapObject.dataType == RegisterDataType.NUMERIC):
                        if(registerMapObject.count == 1):
                            returnValue = response.registers[0]
                        else:
                            decoder = BinaryPayloadDecoder.fromRegisters(response.registers, wordorder=Endian.BIG, byteorder=Endian.BIG)
                            #if(registerMapObject.count == 2):
                            if(registerMapObject.count == 2):
                                returnValue = decoder.decode_32bit_uint()
                            elif(registerMapObject.count == 4):
                                returnValue = decoder.decode_64bit_uint()
                            else:
                                logging.error("Unsupported decoding format {0}".format(registerMapObject.count))
                    elif(registerMapObject.dataType == RegisterDataType.STRING):
                        decoder = BinaryPayloadDecoder.fromRegisters(response.registers, wordorder=Endian.BIG, byteorder=Endian.BIG)
                        returnValue = decoder.decode_string(registerMapObject.count)  
                        returnValue = returnValue.decode('utf-8').replace('\0', '')
                    elif(registerMapObject.dataType == RegisterDataType.VAL_D2FP100):
                        if(registerMapObject.count == 1):
                            
                            fp100value = response.registers[0]
                            if fp100value > 0xF000: #most likely negative
                                fp100value = 0xFFFF - fp100value
                                fp100value = fp100value * -1
                                logging.debug("Detected abnormal high value {0} {1} {2}".format(fp100value, 0xFFFF - fp100value, 0xF000))
        
                            returnValue = float(fp100value / 100)
                            logging.debug("Registers: {0} {1} {2}".format(response.registers[0], float(0x7FFF), 0x7FFF))
                            logging.debug("Found {0} -> /100 = {1}".format(fp100value, returnValue))
                        else:
                            logging.error("Unsupported count for Holding Registers {0} D2_FP100".format(registerMapObject.count))

                    else:
                        logging.error("No support yet for Input Register with {0} type".format(registerMapObject.regType))
                else:
                    logging.debug("Error while reading register {0} type INPUT".format(address)) #not really error, could be register is not used. just return error
                    returnValue = None
                #logging.error("Read InputReg complete {0}".format(returnValue))
            elif(registerMapObject.regType == RegisterType.DISCRETE_INPUT):
                logging.error("Discrete Input type {0} - {1}".format(address, registerMapObject.count))
            
            elif(registerMapObject.regType == RegisterType.HOLDING_REGISTER):
                #logging.error("Holding Register type {0} - {1}".format(registerMapObject.address, registerMapObject.count))
                response = self.client.read_holding_registers(address, registerMapObject.count, slave=self.slaveId)
                if not response.isError(): 
                    if(registerMapObject.dataType == RegisterDataType.NUMERIC):
                        if(registerMapObject.count == 1):
                            returnValue = response.registers[0]
                        else:
                            logging.error("Unsupported count for Holding Registers {0} Numeric".format(registerMapObject.count))
                    elif(registerMapObject.dataType == RegisterDataType.STRING):
                        decoder = BinaryPayloadDecoder.fromRegisters(response.registers, wordorder=Endian.BIG, byteorder=Endian.BIG)
                        returnValue =  decoder.decode_string(registerMapObject.count)
                        returnValue = returnValue.decode('utf-8').replace('\0', '')
                    elif(registerMapObject.dataType == RegisterDataType.VAL_D2FP100):
                        if(registerMapObject.count == 1):
                            fp100value = response.registers[0]
                            if fp100value > 0xF000: #most likely negative
                                fp100value = 0xFFFF - fp100value
                                fp100value = fp100value * -1
                                logging.debug("Detected abnormal high value {0} {1} {2}".format(fp100value, 0xFFFF - fp100value, 0xF000))
        
                            returnValue = fp100value / 100
                            logging.info("Found {0} -> /100 = {1}".format(fp100value, returnValue))
                        else:
                            logging.error("Unsupported count for Holding Registers {0} D2_FP100".format(registerMapObject.count))
                else:
                    logging.debug("Error while reading register") #not really error, could be register is not used. just return error
                    returnValue = None
        except Exception as e:
            logging.exception("error occured ==>  {0}".format(e))
            return None
        return returnValue


class Singleton:  # pylint: disable=too-few-public-methods
    """Singleton base class.
    https://mail.python.org/pipermail/python-list/2007-July/450681.html
    """

    def __new__(cls, *args, **kwargs):  # pylint: disable=unused-argument
        """Create a new instance."""
        if "_inst" not in vars(cls):
            cls._inst = object.__new__(cls)
        return cls._inst

class Defaults(Singleton):
    TcpPort = 502
    SlaveId = 1
    SerialNumberLength = 14
    SerialNumberPrefixLen = 4
    SerialNumberPrefix = 1530
    MaxNumberOfRooms = 24
    MaxNumberOfItcs = 2
    InvalidFP2_100 = 327.67
    BaudRate = 19200


import io
import os
import re

from setuptools import find_packages
from setuptools import setup

def read(filename):
    filename = os.path.join(os.path.dirname(__file__), filename)
    text_type = type(u"")
    with io.open(filename, mode="r", encoding='utf-8') as fd:
        return re.sub(text_type(r':[a-z]+:`~?(.*?)`'), text_type(r'``\1``'), fd.read())

setup(
    name="WavinSentioModbus",
    url="https://github.com/wavingroup/WavinSentioModbus",
    license='MIT',

    author="Wavin T&I",
    author_email="support@wavin.com",

    description="Python API for interfacing with Wavin CCU-208",
    
    long_description=read("Readme.md"),
    long_description_content_type="text/markdown",

    packages=['WavinSentioModbus'],

    version="0.6.3",

    install_requires=['pyserial>=3.5', 'pymodbus>=3.6.0'],

    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
    ],
)


from enum import Enum



class RegisterType(Enum):
    DISCRETE_INPUT = 0
    INPUT_REGISTER = 1
    HOLDING_REGISTER = 2
    INPUT_REGISTER_MULT = 10
    
class RegisterDataType(Enum):
    NUMERIC = 0
    STRING = 1
    VAL_D2FP100 = 10

class RegisterObjectType(Enum):
    GENERIC = 0
    ROOM = 1
    DEHUMIDIFIERS = 10
    ITC_CIRCUIT = 20


class RegisterRepresentation:

    def __init__(
            self, 
            _regType: RegisterType,
            _address: int,
            _count: int,
            _dataType: RegisterDataType = RegisterDataType.NUMERIC,
            _objectType: RegisterObjectType = RegisterObjectType.GENERIC
    ):
        self.regType = _regType
        self.address = _address
        self.count = _count
        self.dataType = _dataType
        self.objectType = _objectType
        pass

    def __str__(self):
        return "Address:{0}, Type:{1}, Count:{2}, DataType:{3}, ObjectType:{4}".format(self.address, self.regType, self.count, self.dataType, self.objectType)

class SentioRegisterMap:
    DeviceType          = RegisterRepresentation(RegisterType.INPUT_REGISTER, 10, 1)
    DeviceHwVersion     = RegisterRepresentation(RegisterType.INPUT_REGISTER, 11, 1)
    DeviceSwVersion     = RegisterRepresentation(RegisterType.INPUT_REGISTER, 12, 1)
    DeviceSwVersionMinor= RegisterRepresentation(RegisterType.INPUT_REGISTER, 13, 1)
    DeviceSerialNrPrefix= RegisterRepresentation(RegisterType.INPUT_REGISTER, 14, 1)
    DeviceSerialNumber  = RegisterRepresentation(RegisterType.INPUT_REGISTER, 15, 2)
    ModbusMode          = RegisterRepresentation(RegisterType.HOLDING_REGISTER, 5, 1)

    HCSourceState       = RegisterRepresentation(RegisterType.INPUT_REGISTER, 8101, 1)
    #Room Registers
    RoomTempSetpoint    = RegisterRepresentation(RegisterType.INPUT_REGISTER, 101, 1,  _objectType = RegisterObjectType.ROOM) #101, 201, 301 - 2401
    RoomAirTempActual   = RegisterRepresentation(RegisterType.INPUT_REGISTER, 104, 1,  _objectType = RegisterObjectType.ROOM) #101, 201, 301 - 2401
    RoomFloorTempActual = RegisterRepresentation(RegisterType.INPUT_REGISTER, 105, 1,  _objectType = RegisterObjectType.ROOM) #101, 201, 301 - 2401
    RoomRelHumidity     = RegisterRepresentation(RegisterType.INPUT_REGISTER, 106, 1,  _objectType = RegisterObjectType.ROOM) #101, 201, 301 - 2401
    RoomCalcDewPoint    = RegisterRepresentation(RegisterType.INPUT_REGISTER, 107, 1,  _objectType = RegisterObjectType.ROOM) #101, 201, 301 - 2401
    RoomHeatingState    = RegisterRepresentation(RegisterType.INPUT_REGISTER, 102, 1,  _objectType = RegisterObjectType.ROOM) #101, 201, 301 - 2401
    RoomBlockingReason  = RegisterRepresentation(RegisterType.INPUT_REGISTER, 103, 1, _objectType = RegisterObjectType.ROOM)

    RoomName            = RegisterRepresentation(RegisterType.HOLDING_REGISTER, 101, 16, _dataType = RegisterDataType.STRING, _objectType = RegisterObjectType.ROOM)
    RoomModeOverride    = RegisterRepresentation(RegisterType.HOLDING_REGISTER, 118, 1, _dataType = RegisterDataType.NUMERIC, _objectType = RegisterObjectType.ROOM)
    RoomTempSetpChange  = RegisterRepresentation(RegisterType.HOLDING_REGISTER, 119, 1, _dataType = RegisterDataType.VAL_D2FP100, _objectType = RegisterObjectType.ROOM)
    RoomTempSetpVacation= RegisterRepresentation(RegisterType.HOLDING_REGISTER, 121, 1, _dataType = RegisterDataType.VAL_D2FP100, _objectType = RegisterObjectType.ROOM)
    RoomTempSetpStandby = RegisterRepresentation(RegisterType.HOLDING_REGISTER, 122, 1, _dataType = RegisterDataType.VAL_D2FP100, _objectType = RegisterObjectType.ROOM)
    RoomMode            = RegisterRepresentation(RegisterType.HOLDING_REGISTER, 117, 1, _objectType = RegisterObjectType.ROOM)
    RoomPresetMode      = RegisterRepresentation(RegisterType.HOLDING_REGISTER, 135, 1, _objectType = RegisterObjectType.ROOM)
    

    OutdoorTemperature  = RegisterRepresentation(RegisterType.INPUT_REGISTER,   3301, 1, _dataType = RegisterDataType.VAL_D2FP100)
    
    #CMV Registers
    CMVDeviceName       = RegisterRepresentation(RegisterType.HOLDING_REGISTER, 61001, 32, RegisterDataType.STRING)
    CMVVentilationState = RegisterRepresentation(RegisterType.INPUT_REGISTER, 61023, 1)

    #DHW
    DHWDeviceName       = RegisterRepresentation(RegisterType.HOLDING_REGISTER, 6501, 32, RegisterDataType.STRING)

    #Dehumidifier
    DehumDeviceName     = RegisterRepresentation(RegisterType.HOLDING_REGISTER, 65001, 32, RegisterDataType.STRING, _objectType = RegisterObjectType.DEHUMIDIFIERS)

    #ITC
    Itc_Name                = RegisterRepresentation(RegisterType.HOLDING_REGISTER, 7301, 32, RegisterDataType.STRING, _objectType = RegisterObjectType.ITC_CIRCUIT)
    Itc_State               = RegisterRepresentation(RegisterType.INPUT_REGISTER, 7301, 1, RegisterDataType.NUMERIC, _objectType = RegisterObjectType.ITC_CIRCUIT)
    Itc_PumpState           = RegisterRepresentation(RegisterType.INPUT_REGISTER, 7304, 1, RegisterDataType.NUMERIC, _objectType = RegisterObjectType.ITC_CIRCUIT)
    Itc_InletTempMeasured   = RegisterRepresentation(RegisterType.INPUT_REGISTER, 7305, 1, RegisterDataType.VAL_D2FP100, _objectType = RegisterObjectType.ITC_CIRCUIT)
    Itc_InletTempDesired    = RegisterRepresentation(RegisterType.INPUT_REGISTER, 7306, 1, RegisterDataType.VAL_D2FP100, _objectType = RegisterObjectType.ITC_CIRCUIT)
    Itc_ReturnTemperature   = RegisterRepresentation(RegisterType.INPUT_REGISTER, 7307, 1, RegisterDataType.VAL_D2FP100, _objectType = RegisterObjectType.ITC_CIRCUIT)
    Itc_SupplierTemperature = RegisterRepresentation(RegisterType.INPUT_REGISTER, 7308, 1, RegisterDataType.VAL_D2FP100, _objectType = RegisterObjectType.ITC_CIRCUIT)
    



import sys
import os
import logging

logging.basicConfig(format='[%(levelname)8s] [%(asctime)s] | %(message)s', level=logging.INFO, datefmt='%m/%d/%Y %H:%M:%S')


#add original module from source to sys path to use it here.
SCRIPT_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), '..')
logging.info("DIR IS {0}".format(SCRIPT_DIR))
sys.path.append(SCRIPT_DIR)


from WavinSentioModbus import SentioApi
from WavinSentioModbus.SentioTypes import *
import time



logging.info("Starting Tests")

class TestClass:
    def init(self):
        returnValue = 0
        logging.debug("Initialize")
        '''
            def __init__(self, 
        _ModbusType:ModbusType = ModbusType.MODBUS_TCPIP,
        _host: str = "",
        _baudrate: int = Defaults.BaudRate,
        _slaveId: int = Defaults.SlaveId,
        _port: int = Defaults.TcpPort,
        _loglevel = logging.ERROR,
        _parity = 'E',
        _stopbits = 1
    ):
        '''
        self.sentio_api = SentioApi.SentioModbus(SentioApi.ModbusType.MODBUS_TCPIP, "192.168.188.14", _port=512)
        #self.sentio_api = SentioApi.SentioModbus("10.31.229.59", SentioApi.ModbusType.MODBUS_TCPIP)
        #self.sentio_api = SentioApi.SentioModbus(SentioApi.ModbusType.MODBUS_RTU, "/dev/ttyS5", 19200, 1, _loglevel=logging.DEBUG)
        #self.sentio_api = SentioApi.SentioModbus("192.168.188.14")
        if self.sentio_api.connect() == 0:
            #logging.error("---- Initializing device data start")
            if self.sentio_api.initialize() == 0:
                logging.debug("Succesfully initialized Sentio device")
            else:
                returnValue = -1
        else:
            logging.error("Failed to connect!")
            returnValue = -1
        return returnValue


    def detectGlobalPeripherals(self):
        status = self.sentio_api.detectDHW()
        if status:
            logging.info("DHW device active")
        else:
            logging.info("DHW device not active")
        status = self.sentio_api.detectCMV()
        if status:
            logging.info("CMV device active")
        else:
            logging.info("CMV device not active")
        status = self.sentio_api.detectDehumidifiers()
        i = 0
        for s in status:
            i = i + 1
            if s == True:
                logging.info("Dehumidification {0} device active".format(i))
            else:
                logging.info("Dehumidification {0} device inactive".format(i))
        
    def readData(self):
        logging.info("DeviceType           = {0}".format(self.sentio_api._sentioData.device_type))
        logging.info("SerialNumber         = {0}".format(self.sentio_api._sentioData.serial_number))
        logging.info("FW Major             = {0}".format(self.sentio_api._sentioData.firmware_version_major))
        logging.info("FW Minor             = {0}".format(self.sentio_api._sentioData.firmware_version_minor))
        #self.sentio_api.readCMVDeviceData()

    def cleanup(self):
        self.sentio_api.disconnect()
    
    def updateData(self):
        self.sentio_api.updateData()

    def showRooms(self):
        logging.info("---------------- ROOMS:  ----------")
        #self.sentio_api.updateRoomData()
        rooms = self.sentio_api.availableRooms
        #logging.info("-- available rooms: {0}".format(rooms))
        for room in rooms:
            logging.info("-- {0}".format(room))
            logging.info("-- Mode {0}".format(room.getRoomMode()))
            logging.info("-- Setpoint {0} °C".format(room.getRoomSetpoint()))
            if room.getRoomActualTemperature() != None:
                logging.info("-- CurrTemp {0} °C".format(room.getRoomActualTemperature()))
            if room.getRoomRelativeHumidity() != None:   
                logging.info("-- RelHumid {0}%".format(room.getRoomRelativeHumidity()))
            if room.getRoomFloorTemperature() != None:
                logging.info("-- FloorTmp {0} °C".format(room.getRoomFloorTemperature()))
            if room.getRoomCalculatedDewPoint() != None:
                logging.info("-- DewPoint {0} °C".format(room.getRoomCalculatedDewPoint()))
            if room.getRoomHeatingState() != None:
                logging.info("-- HeatingState = {0}".format(room.getRoomHeatingState()))
            if room.roomBlockingMode != None:
                logging.info("-- Blocking reason = {0}".format(room.roomBlockingMode))

    def getRoom(self, roomIndex):
        rooms = self.sentio_api.getRooms()
        for room in rooms:
            if room.index == roomIndex:
                return room
        return None

    def showItcCircuits(self):
        logging.info("---------------- ITC Circuits:  ----------")
        for itcCircuit in self.sentio_api.availableItcs:
            logging.info("-- {0}".format(itcCircuit.name))
            logging.info("-- Index      {0}".format(itcCircuit.index))
            logging.info("-- State      {0}".format(itcCircuit._state))
            logging.info("-- PumpState  {0}".format(itcCircuit._pumpState))
            if itcCircuit._inletMeasured:
                logging.info("-- InletTemp  {0} °C".format(itcCircuit._inletMeasured))
            if itcCircuit._inletDesired:
                logging.info("-- InletDes   {0} °C".format(itcCircuit._inletDesired))
            if itcCircuit._returnTemp:
                logging.info("-- ReturnTemp {0} °C".format(itcCircuit._returnTemp))
            if itcCircuit._supplierTemp:
                logging.info("-- SupplierTmp{0} °C".format(itcCircuit._supplierTemp))


    def getHCState(self):
        logging.info("Main HC Source state {0}".format(self.sentio_api._sentioData.hc_source_state))

    def getoutdoorTemp(self):
        logging.info("Outdoor temp = {0} °C".format(self.sentio_api._sentioData.outdoor_temperature))

    def getRoomHeatingState(self, roomIndex):
        heatingState = self.sentio_api.getRoomHeatingState(roomIndex)
        logging.info("Room {0} state {1}".format(roomIndex, heatingState))
        return heatingState

    def getRoomMode(self, roomIndex):
        roomMode = self.sentio_api.getRoomMode(roomIndex)
        logging.info("Room {0} state {1}".format(roomIndex, roomMode))
        return roomMode

    def setRoomToSchedule(self, roomIndex):
        self.sentio_api.setRoomMode(roomIndex, SentioRoomMode.SCHEDULE)
        pass
    
    def setRoomToManual(self, roomIndex):
        self.sentio_api.setRoomMode(roomIndex, SentioRoomMode.MANUAL)
        pass

    def setRoomTemperature(self, roomIndex, temperatureSetpoint):
        self.sentio_api.setRoomSetpoint(roomIndex, temperatureSetpoint)



#Execute Tests
testInstance = TestClass()
assert testInstance.init() == 0, "Failed to connect"
testInstance.readData()
testInstance.detectGlobalPeripherals()

testInstance.updateData()
#testInstance.showRooms()

#Show rooms and outdoor temp
testInstance.getoutdoorTemp()
#testInstance.showRooms()

#firstRoom = testInstance.getRoom(0)
#if firstRoom != None:
#    #firstRoom.setRoomPreset(SentioRoomPreset.RP_COMFORT)
#    firstRoom.setRoomMode(SentioRoomMode.MANUAL)
#    setPoint = firstRoom.getRoomSetpoint()
#    testInstance.sentio_api.setRoomSetpoint(0, 21.5)
#    
#    roomMode = firstRoom.getRoomMode()
#    print("RoomMode = {0} - {1}".format(roomMode, setPoint))
#

testInstance.showItcCircuits()
#testInstance.updateData()

testInstance.showRooms()
testInstance.getHCState()
'''
roomToSet = 0
testInstance.setRoomToSchedule(roomToSet)
assert testInstance.getRoomMode(roomToSet) == SentioRoomMode.SCHEDULE, "ERROR -  Failing to set to schedule"
testInstance.showRooms()
testInstance.setRoomTemperature(roomToSet, 19.5)

testInstance.setRoomToManual(roomToSet)
time.sleep(0.2)
assert testInstance.getRoomMode(roomToSet) == SentioRoomMode.MANUAL, "ERROR -  Failing to set to Manual"
testInstance.showRooms()

#set back
logging.info("========= CLEANUP ==============")
testInstance.setRoomToSchedule(3)
testInstance.setRoomToManual(0)
testInstance.showRooms()

'''
#cleanup
testInstance.cleanup()